from ewah.hooks import hook_class_names


def get_provider_info():
    return {
        "package-name": "ewah",
        "name": "ewah",
        "description": "",
        "hook-class-names": hook_class_names,
        "versions": ["1.0.0"],
    }
